@@ -2,7 +2,7 @@
 #
 # dbginfo.sh - Tool to collect runtime, configuration, and trace information
 #
-# Copyright IBM Corp. 2002, 2023
+# Copyright IBM Corp. 2002, 2024
 #
 # s390-tools is free software; you can redistribute it and/or modify
 # it under the terms of the MIT license. See LICENSE for details.
@@ -15,6 +15,7 @@ export LC_ALL
 ########################################
 # Global used variables
 readonly SCRIPTNAME="${0##*/}" # general name of this script
+readonly FULLPATHSCRIPT="$(readlink -f "${0}")"
 #
 readonly DATETIME="$(date +%Y-%m-%d-%H-%M-%S 2>/dev/null)"
 readonly DOCKER=$(if type docker >/dev/null 2>&1; then echo "YES"; else echo "NO"; fi)
@@ -51,6 +52,14 @@ if test "x${PROCESSORVERSION}" = "xFF" || test "x${PROCESSORVERSION}" = "xff"; t
 else
 	RUNTIME_ENVIRONMENT="LPAR"
 fi
+# check for Dynamic Partition Mode
+readonly DPM_UUID=$(grep "LPAR UUID" /proc/sysinfo | \
+	sed 's/.*:[[:space:]]*\([[:graph:]]*\).*/\1/g')
+if test "${#DPM_UUID}" -eq 0; then
+	DPM_MODE="NO"
+else
+	DPM_MODE="YES"
+fi
 readonly SYSTEMHOSTNAME="$(hostname -s 2>/dev/null)" # hostname of system being analysed
 readonly TERMINAL="$(tty 2>/dev/null)"
 # timeout seconds TOS / kill timeout TOKS
@@ -71,8 +80,8 @@ paramWORKDIR_BASE="/tmp"  # initial default path
 # print dbginfo.sh version info
 print_version() {
 	cat <<EOF
-${SCRIPTNAME}: Debug information script version 2.29.0-build-20240423
-Copyright IBM Corp. 2002, 2023
+${SCRIPTNAME}: Debug information script version 2.31.0-build-20240809
+Copyright IBM Corp. 2002, 2024
 EOF
 }
 
@@ -155,7 +164,7 @@ print_check() {
 	cat <<EOF
 
 Hardware platform     = ${HW}
-Runtime environment   = ${RUNTIME_ENVIRONMENT}
+Runtime environment   = ${RUNTIME_ENVIRONMENT} - DPM: ${DPM_MODE}
 $(cat /proc/sysinfo | grep 'Name')
 Kernel version        = ${KERNEL_INFO}
 OS version / distro   = ${OS_NAME}
@@ -262,6 +271,7 @@ readonly OUTPUT_FILE_OVS="${WORKPATH}network.out"
 readonly OUTPUT_FILE_ISW="${WORKPATH}installed_sw.out"
 readonly OUTPUT_FILE_TC="${WORKPATH}network.out"
 readonly OUTPUT_FILE_VMCMD="${WORKPATH}zvm_runtime.out"
+readonly OUTPUT_FILE_UDEVDB="${WORKPATH}udevdb.out"
 # Base file names for different output files - no extension !
 readonly OUTPUT_FILE_KUBERNETES="${WORKPATH}kubernetes"
 readonly OUTPUT_FILE_OSAOAT="${WORKPATH}network"
@@ -273,6 +283,7 @@ ALL_STEPS="\
   collect_sysfs\
   collect_procfs\
   collect_configfiles\
+  collect_initrd_configfiles\
   collect_cmdsout\
   collect_hyptop\
   collect_vmcmdsout\
@@ -380,14 +391,19 @@ CONFIGFILES="\
   /boot/loader/entries/*.conf\
   /boot/zipl/active_devices.txt\
   /boot/zipl/config\
+  /conf/\
   /etc/*.conf\
   /etc/*release\
   /etc/anacrontab\
   /etc/apparmor.d\
   /etc/auto.*\
+  /etc/cmdline\
+  /etc/cmdline.d\
+  /etc/conf.d\
   /etc/cron.*\
   /etc/crontab\
   /etc/crypttab\
+  /etc/dasd.*\
   /etc/default\
   /etc/depmod.d\
   /etc/dnf/dnf.conf\
@@ -400,6 +416,7 @@ CONFIGFILES="\
   /etc/hosts*\
   /etc/iscsi\
   /etc/inittab\
+  /etc/kdump\
   /etc/libvirt\
   /etc/logrotate.d\
   /etc/lvm\
@@ -427,18 +444,22 @@ CONFIGFILES="\
   /etc/systemd\
   /etc/udev*\
   /etc/xinet.d\
+  /etc/zfcp.*\
   /lib/modprobe.d\
   $(find /lib/modules -name modules.dep 2>/dev/null)\
   /lib/systemd/system/docker.service\
+  /lib/udev/rules.d\
   /usr/lib/modprobe.d\
   /usr/lib/modules-load.d/*.conf\
-  /usr/lib/systemd/system\
+  /usr/lib/systemd/system/\
   /usr/local/lib/modprobe.d\
   /usr/local/lib/modules-load.d/*.conf\
   /run/modprobe.d\
   /run/modules-load.d/*.conf\
   /run/udev/chreiplzfcpmp-ipl-volume-id\
   /run/udev/rules.d\
+  /run/zdev_id.env\
+  /run/zdev.*\
   "
 
 ########################################
@@ -555,6 +576,7 @@ CMDS="${CMDS}\
    >> '${OUTPUT_FILE_JOURNALCTL}'\
   :lsof >> '${OUTPUT_FILE_LSOF}'\
   :rpm -qa | sort >> '${OUTPUT_FILE_ISW}'\
+  :udevadm info --export-db >> '${OUTPUT_FILE_UDEVDB}'\
   "
 
 ########################################
@@ -833,6 +855,7 @@ collect_sysfs() {
 		    -x '*/page_idle/bitmap*' \
 		    -x '*/tracing/per_cpu/*' \
 		    -x '*/tracing/free_buffer' \
+		    -x '*/devices/css[0-9]*/[0-9]*.*.*/[0-9]*.*.*/host[0-9]*/rport-[0-9]*:*-*/target[0-9]*:*:*/[0-9]*:*:*:*/block/sd*/mq/0/cpu[0-9]*' \
 		    --ignore-failed-read -J 1 -T 5 2>>${OUTPUT_FILE_SYSFS}.err
 		rc=$?
 	else
@@ -860,6 +883,7 @@ collect_sysfs() {
 		    -a -not \( -path '*page_idle*' -a -name 'bitmap*' \) \
 		    -a -not -path '*tracing/per_cpu*' \
 		    -a -not -path '*/tracing/free_buffer' \
+		    -a -not -path '*/devices/css[0-9]*/[0-9]*.*.*/[0-9]*.*.*/host[0-9]*/rport-[0-9]*:*-*/target[0-9]*:*:*/[0-9]*:*:*:*/block/sd*/mq/0/cpu[0-9]*' \
 		    2>/dev/null | while IFS= read -r file_name; do
 			echo " ${file_name}"
 			if ! dd if="${file_name}" status=noxfer iflag=nonblock \
@@ -898,13 +922,97 @@ collect_configfiles() {
 	done
 }
 
+########################################
+collect_initrd_configfiles() {
+	local file_name
+	local rd rdbase rdreldir rdextract mypopd noglob mycat patterns=""
+
+	pr_collect "initrd config files"
+	cd "${WORKPATH}" || return
+	mypopd="$OLDPWD" # (pushd alternative)
+	# prevent expansion of glob special characters such as '*'
+	noglob=$(set +o | grep -F noglob)
+	set -o noglob
+	# prepare relative configfiles paths
+	for file_name in ${CONFIGFILES}; do
+	    # strip leading slash to make path relative as in initrd
+	    file_name=${file_name#/}
+	    # in contrast to call_collect_file(), cpio extract with patterns
+	    # does not recurse, but cpio wildcards match slash, so add suffix;
+	    # otherwise e.g. /lib/udev/rules.d or /etc/multipath only extracts
+	    # just that directory without any of its content
+	    file_name="${file_name}*"
+	    patterns="$patterns ${file_name}"
+	done
+	patterns="$patterns squash-root.img"
+	$noglob
+	# iterate over initrds
+	for rd in /boot/initr* /boot/zipl/initr* /var/lib/kdump/initr*; do
+	    [ -f "$rd" ] || continue # (nullglob alternative)
+	    rdbase=${rd##*/} # basename
+	    rdreldir=${rd%/*} # dirname
+	    rdreldir=${rdreldir#/} # strip leading slash to make path relative
+	    mkdir -p "${rdreldir}"
+	    cd "${rdreldir}" || break # (pushd alternative)
+	    if type lsinitrd >/dev/null 2>&1; then
+		lsinitrd "${rd}" > "${rdbase}.lsinitrd"
+	    fi
+	    if type lsinitramfs >/dev/null 2>&1; then
+		lsinitramfs -l "${rd}" > "${rdbase}.lsinitramfs"
+	    fi
+	    cd - || break # change to $OLDPWD (popd alternative)
+	    rdextract=${rd#/}
+	    mkdir -p "${rdextract}"
+	    cd "${rdextract}" || break # (pushd alternative)
+	    case $(file "${rd}") in
+		*cpio*) mycat="cat" ;;
+		*gzip*) mycat="zcat" ;;
+		*XZ*) mycat="xzcat" ;;
+		*Zstandard*) mycat="zstdcat" ;;
+		*bzip2*) mycat="bzcat" ;;
+		*LZMA*) mycat="lzcat" ;;
+		*)
+		    # change to $OLDPWD (popd alternative)
+		    cd - || { cd "$mypopd" || return; return; }
+		    continue
+		    ;;
+	    esac
+	    # prevent expansion of glob special characters such as '*'
+	    noglob=$(set +o | grep -F noglob)
+	    set -o noglob
+	    # we expect no early-initrd stuff on s390 so extract directly.
+	    # globbing is turned off and we want word split on $patterns:
+	    # shellcheck disable=SC2086
+	    "$mycat" "${rd}" | cpio --extract --no-absolute-filenames \
+				    --make-directories \
+				    --preserve-modification-time $patterns
+	    $noglob
+	    if [ ! -f "squash-root.img" ]; then
+		cd - || break # change to $OLDPWD (popd alternative)
+		continue
+	    fi
+	    # prevent expansion of glob special characters such as '*'
+	    noglob=$(set +o | grep -F noglob)
+	    set -o noglob
+	    # globbing is turned off and we want word split on $patterns:
+	    # shellcheck disable=SC2086
+	    unsquashfs -no-progress -dest "squashfs-root" "squash-root.img" \
+		       $patterns
+	    $noglob
+	    rm -f "squash-root.img"
+	    cd - || break # change to $OLDPWD (popd alternative)
+	done
+	cd "$mypopd" || return
+}
+
 ########################################
 collect_osaoat() {
 	local network_devices
 	local network_device
 
 	network_devices=$(lsqeth 2>/dev/null | grep "Device name" \
-		     | sed 's/D.*:[[:space:]]*\([^[:space:]]*\)[[:space:]]\+/\1/g')
+		     | sed 's/D.*:[[:space:]]*\([^[:space:]]*\)[[:space:]]\+/\1/g' \
+		     | sed 's/[()]//g' )
 	if type qethqoat >/dev/null; then
 		if test -n "${network_devices}"; then
 			pr_collect_output "osa oat"
@@ -954,6 +1062,8 @@ collect_ethtool() {
 					"${OUTPUT_FILE_ETHTOOL}"
 				call_run_command "ethtool -T ${network_device}" \
 					"${OUTPUT_FILE_ETHTOOL}"
+				call_run_command "ethtool --module-info ${network_device}" \
+					"${OUTPUT_FILE_ETHTOOL}"
 			done
 		else
 			pr_skip "ethtool: no devices"
@@ -1310,10 +1420,10 @@ environment_setup() {
 create_package() {
 	local rc_tar
 	pr_syslog_stdout ${step_num} "Finalizing: Creating archive with collected data"
-	cd "${WORKDIR_BASE}"
 	# get a copy of the script used
-	cp -p $0 "${WORKDIR_CURRENT}"
-
+	cp -p "${FULLPATHSCRIPT}" "${WORKPATH}"
+	# create the archive
+	cd "${WORKDIR_BASE}"
 	touch "${WORKARCHIVE}"
 	chmod 0600 "${WORKARCHIVE}"
 	tar -czf "${WORKARCHIVE}" "${WORKDIR_CURRENT}"
@@ -1389,7 +1499,7 @@ pr_log_stdout() {
 pr_syslog_stdout() {
 	echo "$@" >&8
 	echo
-	echo "$(date +%H:%M:%S.%N) - $@"
+	echo "$(date +%H:%M:%S.%N) - " "$@"
 	logger -t "${SCRIPTNAME}" "$@"
 }
 
@@ -1402,13 +1512,13 @@ pr_collect_output() {
 ########################################
 # print "Collecting ..." like fs
 pr_collect() {
-	pr_syslog_stdout ${step_num} "Collecting" $@
+	pr_syslog_stdout ${step_num} "Collecting" "$@"
 }
 
 ########################################
 # print "Skipping ..." info with reason
 pr_skip() {
-	pr_syslog_stdout ${step_num} "Skip" $@
+	pr_syslog_stdout ${step_num} "Skip" "$@"
 }
 
 ###############################################################################
@@ -1428,7 +1538,7 @@ trap emergency_exit SIGHUP SIGINT SIGTERM
 
 pr_log_stdout ""
 pr_log_stdout "Hardware platform     = ${HW}"
-pr_log_stdout "Runtime environment   = ${RUNTIME_ENVIRONMENT}"
+pr_log_stdout "Runtime environment   = ${RUNTIME_ENVIRONMENT} - DPM: ${DPM_MODE}"
 pr_log_stdout "Kernel version        = ${KERNEL_INFO} (${KERNEL_BASE})"
 pr_log_stdout "OS version / distro   = ${OS_NAME}"
 pr_log_stdout "Date and time of info = ${DATETIME}"
